from typing import Union
import matplotlib.pyplot as plt
import numpy as np
from ..DynTrack import DynTrack
from matplotlib.axes import Axes


def tracks(
    DT: DynTrack,
    figsize: tuple = (7, 4),
    ax: Union[Axes, None] = None,
    show: bool = True,
    **kwargs
):
    """\
    Plotting all single tracks.

    Parameters
    ----------
    DT
        A :class:`dyntrack.DynTrack` object.
    figsize
        Figure size.
    ax
        A matplotlib axes object.
    show
        Show the plot, do not return axis.
    **kwargs
        Arguments passed to :func:`matplotlib.pyplot.plot`.

    Returns
    -------
    :class:`matplotlib.axes.Axes` if `show=True`

    """

    if ax is None:
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(111)
        fig.set_tight_layout(True)

    ax.set_aspect("equal")
    if DT.img is not None:
        ax.imshow(DT.img, origin="lower")

    for t in DT.track_data.Parent.unique():
        ax.plot(
            DT.track_data.loc[DT.track_data.Parent == t, "Position X"],
            DT.track_data.loc[DT.track_data.Parent == t, "Position Y"],
            **kwargs
        )

    ax.axis("off")

    if show == False:
        return ax
    else:
        plt.show()



from typing import Union
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy as np
from ..DynTrack import DynTrack
from matplotlib.axes import Axes


def FTLE(
    DT: DynTrack,
    cmap="jet",
    density: float = 2,
    linewidth: float = 0.75,
    arrowsize: float = 1,
    arrowstyle: str = "->",
    color="white",
    figsize=(7, 4),
    show: bool = True,
    kwargs_for_countourf={},
    kwargs_for_streamplot={},
):
    """\
    Plotting counterpart of `tl.FTLE`.

    Parameters
    ----------
    DT
        A :class:`dyntrack.DynTrack` object.
    cmap
        Colormap used by :func:`matplotlib.pyplot.countourf`.
    density
        Density of arrows used by :func:`matplotlib.pyplot.streamplot`.
    linewidth
        Arrow line width used by :func:`matplotlib.pyplot.streamplot`.
    arrowsize
        Arrow size used by :func:`matplotlib.pyplot.streamplot`.
    arrowstyle
        Arrow style used by :func:`matplotlib.pyplot.streamplot`.
    color
        Arrow color used by :func:`matplotlib.pyplot.streamplot`.
    figsize
        Figure size.
    show
        Show the plot, do not return axis.
    **kwargs_for_contourf
        Arguments passed to :func:`matplotlib.pyplot.contourf`.
    **kwargs_for_streamplot
        Arguments passed to :func:`matplotlib.pyplot.streamplot`.

    Returns
    -------
    :class:`matplotlib.axes.Axes` if `show=True`

    """

    fig, (ax, cax) = plt.subplots(
        1, 2, gridspec_kw={"width_ratios": [96, 4], "wspace": 0}
    )
    ax.set_aspect("equal")

    contf = ax.contourf(
        DT.X, DT.Y, DT.ftle, extend="both", cmap=cmap, **kwargs_for_countourf
    )
    ax.streamplot(
        DT.X,
        DT.Y,
        DT.u,
        DT.v,
        density=2,
        linewidth=0.75,
        arrowsize=0.75,
        arrowstyle="->",
        color=color,
        **kwargs_for_streamplot
    )
    ax.yaxis.set_major_locator(plt.NullLocator())
    ax.xaxis.set_major_formatter(plt.NullFormatter())
    ax.axis("off")

    cbar = plt.colorbar(contf, ax=cax, pad=0, fraction=1)
    cbar.set_label("$FTLE$", fontsize=12)
    cax.axis("off")

    if show == False:
        return (ax, cax)
    else:
        plt.show()



from typing import Union, Sequence
import matplotlib.pyplot as plt
from matplotlib.axes import Axes
import numpy as np
import simpleppt
from ..DynTrack import DynTrack


def fit_ppt(
    DT: DynTrack,
    times: Union[Sequence, None] = None,
    figsize: tuple = (7, 4),
    ax: Union[Axes, None] = None,
    show: bool = True,
):
    """\
    Plotting counterpart of `tl.fit_ppt`.

    Display all principal trees to reveal structure in trajectories.

    Parameters
    ----------
    DT
        A :class:`dyntrack.DynTrack` object.
    times
        set of timepoints/frames to plot, by default uses all.
    figsize
        Figure size.
    show
        Show the plot, do not return axis.

    Returns
    -------
    :class:`matplotlib.axes.Axes` if `show=True`

    """

    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)
    fig.set_tight_layout(True)

    tdata = DT.track_data

    times = tdata.Time.unique() if times is None else times
    ax.set_aspect("equal")
    i = 0
    for t in times:
        toproject = tdata.loc[tdata.Time == t, ["Position X", "Position Y"]]
        simpleppt.project_ppt(
            DT.ppts[i],
            toproject,
            ax=ax,
            plot_datapoints=False,
            alpha_nodes=0,
            alpha_seg=0.05,
            show=False,
        )
        i = i + 1

    if DT.img is not None:
        ax.imshow(DT.img, origin="lower")
    ax.axis("off")

    if show == False:
        return ax
    else:
        plt.show()



from typing import Any, Union, Optional  # Meta
from typing import Mapping
import pandas as pd
import numpy as np


class DynTrack:
    """A python object containing the data used for dynamical tracks analysis.

    Parameters
    ----------
    track_data
        DataFrame containing x and z coordinates, the track ID and the frame/time.
    img
        The background image from the field where particle were tracked can be included.
    X
        x coordinates of the grid vector field.
    Y
        y coordinates of the grid vector field.
    u
        x component of the vectors.
    v
        y component of the vectors.
    ftle
        scalar FTLE values calculated from vector field.
    ppts
        list of principal trees fitted for each frame of the tracking."""

    def __init__(
        self,
        track_data: pd.DataFrame,
        img: Optional[Union[np.ndarray, None]] = None,
        X: Optional[Union[np.ndarray, None]] = None,
        Y: Optional[Union[np.ndarray, None]] = None,
        u: Optional[Union[np.ndarray, None]] = None,
        v: Optional[Union[np.ndarray, None]] = None,
        ftle: Optional[Union[np.ndarray, None]] = None,
        ppts: Optional[Mapping[str, Any]] = None,
    ):
        self.track_data = track_data
        self.img = img
        self.X = X
        self.Y = Y
        self.u = u
        self.v = v
        self.ftle = ftle
        self.ppts = ppts

    def __repr__(self):
        dp = self.track_data.shape[0]
        tr = len(self.track_data.Parent.unique())

        descr = f"DynTrack object with the following data:"
        descr += f"\n    track_data ({tr} tracks, {dp} datapoints)"
        for attr in ["img", "X", "Y", "u", "v", "ftle", "ppts"]:
            dt = getattr(self, attr)
            if (dt is not None) & (attr != "ppts"):
                descr += f"\n    {attr} {dt.shape}"
            if (dt is not None) & (attr == "ppts"):
                descr += f"\n    {attr} ({len(dt)} ppt)"

        return descr



try:
    from importlib_metadata import version
except:
    from importlib.metadata import version
__version__ = version(__name__)
del version

from . import tl
from . import ut
from . import pl
from . import settings
from .DynTrack import DynTrack



"""Logging and Profiling
"""

from . import settings
from sys import stdout
from datetime import datetime
from time import time as get_time
from platform import python_version


_VERBOSITY_LEVELS_FROM_STRINGS = {"error": 0, "warn": 1, "info": 2, "hint": 3}


def info(*args, **kwargs):
    return msg(*args, v="info", **kwargs)


def error(*args, **kwargs):
    args = ("Error:",) + args
    return msg(*args, v="error", **kwargs)


def warn(*args, **kwargs):
    args = ("WARNING:",) + args
    return msg(*args, v="warn", **kwargs)


def hint(*args, **kwargs):
    return msg(*args, v="hint", **kwargs)


def _settings_verbosity_greater_or_equal_than(v):
    if isinstance(settings.verbosity, str):
        settings_v = _VERBOSITY_LEVELS_FROM_STRINGS[settings.verbosity]
    else:
        settings_v = settings.verbosity
    return settings_v >= v


def msg(
    *msg,
    v=None,
    time=False,
    memory=False,
    reset=False,
    end="\n",
    no_indent=False,
    t=None,
    m=None,
    r=None,
):
    """Write message to logging output.
    Log output defaults to standard output but can be set to a file
    by setting `sc.settings.log_file = 'mylogfile.txt'`.
    v : {'error', 'warn', 'info', 'hint'} or int, (default: 4)
        0/'error', 1/'warn', 2/'info', 3/'hint', 4, 5, 6...
    time, t : bool, optional (default: False)
        Print timing information; restart the clock.
    memory, m : bool, optional (default: Faulse)
        Print memory information.
    reset, r : bool, optional (default: False)
        Reset timing and memory measurement. Is automatically reset
        when passing one of ``time`` or ``memory``.
    end : str (default: '\n')
        Same meaning as in builtin ``print()`` function.
    no_indent : bool (default: False)
        Do not indent for ``v >= 4``.
    """
    # variable shortcuts
    if t is not None:
        time = t
    if m is not None:
        memory = m
    if r is not None:
        reset = r
    if v is None:
        v = 4
    if isinstance(v, str):
        v = _VERBOSITY_LEVELS_FROM_STRINGS[v]
    if v == 3:  # insert "--> " before hints
        msg = ("-->",) + msg
    if v >= 4 and not no_indent:
        msg = ("   ",) + msg
    if _settings_verbosity_greater_or_equal_than(v):
        if not time and not memory and len(msg) > 0:
            _write_log(*msg, end=end)
        if reset:
            settings._previous_time = get_time()
        if time:
            elapsed = get_passed_time()
            msg = msg + (f"({_sec_to_str(elapsed)})",)
            _write_log(*msg, end=end)


m = msg


def _write_log(*msg, end="\n"):
    """Write message to log output, ignoring the verbosity level.
    This is the most basic function.
    Parameters
    ----------
    *msg :
        One or more arguments to be formatted as string. Same behavior as print
        function.
    """
    from .settings import logfile

    if logfile == "":
        print(*msg, end=end)
    else:
        out = ""
        for s in msg:
            out += f"{s} "
        with open(logfile, "a") as f:
            f.write(out + end)


def _sec_to_str(t, show_microseconds=False):
    """Format time in seconds.
    Parameters
    ----------
    t : int
        Time in seconds.
    """
    from functools import reduce

    t_str = "%d:%02d:%02d.%02d" % reduce(
        lambda ll, b: divmod(ll[0], b) + ll[1:], [(t * 100,), 100, 60, 60]
    )
    return t_str if show_microseconds else t_str[:-3]


def get_passed_time():
    now = get_time()
    elapsed = now - settings._previous_time
    settings._previous_time = now
    return elapsed


def print_passed_time():
    return _sec_to_str(get_passed_time())


def timeout(func, args=(), timeout_duration=2, default=None, **kwargs):
    """This will spwan a thread and run the given function using the args, kwargs and
    return the given default value if the timeout_duration is exceeded
    """
    import threading

    class InterruptableThread(threading.Thread):
        def __init__(self):
            threading.Thread.__init__(self)
            self.result = default

        def run(self):
            try:
                self.result = func(*args, **kwargs)
            except:
                pass

    it = InterruptableThread()
    it.start()
    it.join(timeout_duration)
    return it.result



import numpy as np
import math
from matplotlib import pyplot
import time
import sys
import numba


@numba.jit
def bilinear_interpolation(X, Y, f, x, y):
    """Returns the approximate value of f(x,y) using bilinear interpolation.

    Arguments
    ---------
    X, Y -- mesh grid.
    f -- the function f that should be an NxN matrix.
    x, y -- coordinates where to compute f(x,y)

    """

    N = np.shape(X[:, 0])[0]

    dx, dy = X[0, 1] - X[0, 0], Y[1, 0] - Y[0, 0]
    x_start, y_start = X[0, 0], Y[0, 0]

    i1, i2 = int((x - x_start) / dx), int((x - x_start) / dx) + 1
    j1, j2 = int((y - y_start) / dy), int((y - y_start) / dy) + 1

    # Take care of boundaries

    # 1. Right boundary

    if i1 >= N - 1 and j1 <= N - 1 and j1 >= 0:
        return f[j1, N - 1]
    if i1 >= N - 1 and j1 <= 0:
        return f[0, N - 1]
    if i1 >= N - 1 and j1 >= N - 1:
        return f[N - 1, N - 1]

    # 2. Left boundary

    if i1 <= 0 and j1 <= N - 1 and j1 >= 0:
        return f[j1, 0]
    if i1 <= 0 and j1 <= 0:
        return f[0, 0]
    if i1 <= 0 and j1 >= N - 1:
        return f[N - 1, 0]

    # 3. Top boundary

    if j1 >= N - 1 and i1 <= N - 1 and i1 >= 0:
        return f[N - 1, i1]
    if j1 >= N - 1 and i1 <= 0:
        return f[N - 1, 0]

    # 3. Bottom boundary

    if j1 <= 0 and i1 <= N - 1 and i1 >= 0:
        return f[0, i1]
    if j1 <= 0 and i1 >= N - 1:
        return f[N - 1, 0]

    x1, x2 = X[j1, i1], X[j2, i2]
    y1, y2 = Y[j1, i1], Y[j2, i2]

    f_interpolated = (
        1
        / (x2 - x1)
        * 1
        / (y2 - y1)
        * (
            f[j1, i1] * (x2 - x) * (y2 - y)
            + f[j1, i2] * (x - x1) * (y2 - y)
            + f[j2, i1] * (x2 - x) * (y - y1)
            + f[j2, i2] * (x - x1) * (y - y1)
        )
    )

    return f_interpolated


@numba.jit
def rk4(X, Y, x, y, f, h, dim):
    """Returns the approximate value of f(x,y) using bilinear interpolation.

    Arguments
    ---------
    X, Y -- mesh grid.
    x, y -- coordinates where to begin the evolution.
    f -- the function f that will be evolved.
    h -- the time step (usually referred to this as dt.)
    dim -- 0 for x and 1 for y.

    """

    k1 = h * bilinear_interpolation(X, Y, f, x, y)
    k2 = h * bilinear_interpolation(X, Y, f, x + 0.5 * h, y + 0.5 * k1)
    k3 = h * bilinear_interpolation(X, Y, f, x + 0.5 * h, y + 0.5 * k2)
    k4 = h * bilinear_interpolation(X, Y, f, x + h, y + k3)

    if dim == 0:
        return x + 1.0 / 6 * k1 + 1.0 / 3 * k2 + 1.0 / 3 * k3 + 1.0 / 6 * k4
    elif dim == 1:
        return y + 1.0 / 6 * k1 + 1.0 / 3 * k2 + 1.0 / 3 * k3 + 1.0 / 6 * k4
    else:
        print("invalid dimension parameter passed to rk4, exiting")
        # sys.exit()


@numba.jit
def integrate(x_y, integration_time, dt, X, Y, u, v):
    xs = x_y[0]
    ys = x_y[1]
    tr_x = xs
    tr_y = ys
    for k in range(0, int(integration_time / dt)):
        xs, ys = rk4(X, Y, xs, ys, u, dt, 0), rk4(X, Y, xs, ys, v, dt, 1)
        tr_x += xs
        tr_y += ys
    return [tr_x, tr_y]


def eigs(xt, yt, xo, yo):
    ftlemat = np.zeros((2, 2))
    ftlemat[0][0] = (xt[1] - xt[0]) / (xo[1] - xo[0])
    ftlemat[1][0] = (yt[1] - yt[0]) / (xo[1] - xo[0])
    ftlemat[0][1] = (xt[3] - xt[2]) / (yo[1] - yo[0])
    ftlemat[1][1] = (yt[3] - yt[2]) / (yo[1] - yo[0])

    if True in np.isnan(ftlemat):
        return "nan"

    ftlemat = np.dot(ftlemat.transpose(), ftlemat)
    w, v = np.linalg.eig(ftlemat)

    return w



from typing import Union, Optional
import matplotlib.image as mpimg
import pandas as pd
import numpy as np
from ..DynTrack import DynTrack
from .. import __path__
import os


def load_data(
    df: Union[pd.DataFrame, str],
    x_col: str,
    y_col: str,
    parent_col: str,
    time_col: str,
    img: Optional[Union[np.ndarray, str]] = None,
):
    """\
    Load data required for dynamical tracks analysis.

    Parameters
    ----------
    df
        Path of a csv file, or a :class:`pandas.DataFrame` object.
    x_col
        Name of the x coordinate column.
    x_col
        Name of the y coordinate column.
    parent_col
        Name of the track ID column.
    time_col
        Name of the time/frame ID column.
    img
        Path of an img file, or a :class:`numpy.ndarray`.

    Returns
    -------
    :class:`dyntrack.DynTrack`
        A DynTrack object
    """

    tdata = pd.read_csv(df) if type(df) is str else df
    tdata = tdata[[x_col, y_col, parent_col, time_col]]
    tdata.columns = ["Position X", "Position Y", "Parent", "Time"]
    img = mpimg.imread(img) if type(img) is str else img

    return DynTrack(tdata, img)


def load_example():
    csv = os.path.join(__path__[0], "data", "tracks.csv")
    img = os.path.join(__path__[0], "data", "background.tiff")
    return load_data(csv, "Position X", "Position Y", "Parent", "Time", img)



from glob import glob
import sys
import os

import subprocess
import tempfile

import pandas as pd
import numpy as np

from .. import settings
from .. import logging as logg
from ..DynTrack import DynTrack


def vector_field(
    DT: DynTrack, gridRes: int = 30, smooth: float = 0.5, copy: float = False
):
    """\
    Generate a grid vector field from track data.

    Parameters
    ----------
    DT
        A :class:`dyntrack.DynTrack` object.
    gridRes
        grid resolution in both horizontal and vertical axis.
    smooth
        Smooth parameter of the vfkm algorithm.
    copy
        Return a copy instead of writing to DT.

    Returns
    -------
    DT : :class:`dyntrack.DynTrack`
        if `copy=True` it returns or else add fields to `DT`:

        `.X`
            x coordinates of the grid.
        `.Y`
            y coordinates of the grid.
        `.u`
            x component of the vectors.
        `.v`
            y component of the vectors.

    """

    DT = DT.copy() if copy else DT

    tdata = DT.track_data

    allparents = np.unique(tdata["Parent"])
    allparents = allparents[~np.isnan(allparents)]
    ldata = list(
        map(
            lambda x: tdata[["Position X", "Position Y", "Time"]].loc[
                tdata.index[
                    tdata["Parent"].isin(
                        [x],
                    )
                ]
            ],
            [par for par in allparents],
        )
    )

    vfkm = settings.vfkm

    logg.info(
        f"Generating grid vector field with a {gridRes}x{gridRes} grid and a smoothing constrain of {smooth}",
        reset=True,
    )
    with tempfile.TemporaryDirectory() as tmp:
        # logg.hint("    Data temporary saved in "+tmp)
        init = [
            min(tdata["Position X"]),
            max(tdata["Position X"]),
            min(tdata["Position Y"]),
            max(tdata["Position Y"]),
            1,
            max(tdata["Time"]),
        ]
        with open(tmp + "/tracks", "w") as f:
            f.writelines(["%s " % item for item in init])
        for ld in ldata:
            ld.to_csv(tmp + "/tracks", mode="a", header=False, sep=" ", index=False)
            pd.DataFrame([0, 0, 0]).transpose().to_csv(
                tmp + "/tracks", header=False, sep=" ", mode="a", index=False
            )

        proc = subprocess.Popen(
            vfkm
            + " "
            + tmp
            + "/tracks "
            + str(gridRes)
            + " 1 "
            + str(smooth)
            + " "
            + tmp,
            stdout=subprocess.PIPE,
            shell=True,
        )
        (out, err) = proc.communicate()

        files_c = [f for f in glob(tmp + "/*curves_r_*", recursive=True)]
        files_v = [f for f in glob(tmp + "/*vf_r_*", recursive=True)]
        curves = pd.read_table(files_c[0], header=None, sep=" ")
        vect = pd.read_table(files_v[0], header=None, sep=" ", skiprows=1)

    gs = int(np.sqrt(vect.shape[0]))

    a = np.linspace(
        min(pd.concat(ldata)["Position X"]), max(pd.concat(ldata)["Position X"]), gs
    )
    b = np.linspace(
        min(pd.concat(ldata)["Position Y"]), max(pd.concat(ldata)["Position Y"]), gs
    )
    out = np.stack([each.ravel(order="C") for each in np.meshgrid(a, b)])

    u = vect[0].values.reshape(gs, gs)
    v = vect[1].values.reshape(gs, gs)
    X, Y = np.meshgrid(a, b)

    DT.X, DT.Y, DT.u, DT.v = X, Y, u, v

    logg.info("    finished", time=True, end=" " if settings.verbosity > 2 else "\n")
    logg.hint(
        "added \n"
        "    .X, x coordinates of the grid\n"
        "    .Y, y coordinates of the grid\n"
        "    .u, x component of the vectors\n"
        "    .v, y component of the vectors"
    )

    return DT if copy else None



import numpy as np
from .. import logging as logg
from .. import settings
from ..utils.FTLE import *
from ..DynTrack import DynTrack
from tqdm import tqdm
import itertools


def get_traj(X, Y, u, v, integration_time, dt, verbose=True):
    """Returns the FTLE particle trajectory, faster approach

    Arguments
    ---------
    x, y -- mesh grid coordinates
    dt -- integral time step
    """

    N = np.shape(X[:, 1])[0]
    x = X[0, :]
    y = Y[:, 0]

    xy = np.array(np.meshgrid(x, y)).T.reshape(-1, 2).tolist()
    xy = list(map(np.array, xy))

    res = [
        integrate(x_y, integration_time, dt, X, Y, u, v)
        for x_y in tqdm(xy, file=sys.stdout, desc="    integrating")
    ]
    res = np.array(res)

    return res[:, 0].reshape(N, N).T, res[:, 1].T.reshape(N, N).T


def get_ftle(traj_x, traj_y, X, Y, integration_time):
    """Returns the FTLE scalar field

    Mostly adapted from Steven's FTLE code (GitHub user stevenliuyi)

    Arguments
    ---------
    traj_x, traj_y -- The trajectories of the FTLE particles
    X, Y -- Meshgrid
    integration_time -- the duration of the integration time

    """

    N = np.shape(X[:, 0])[0]
    ftle = np.zeros((N, N))

    for i, j in itertools.product(range(0, N), range(0, N)):
        # index 0:left, 1:right, 2:down, 3:up
        xt = np.zeros(4)
        yt = np.zeros(4)
        xo = np.zeros(2)
        yo = np.zeros(2)

        if i == 0:
            xt[0] = traj_x[j][i]
            xt[1] = traj_x[j][i + 1]
            yt[0] = traj_y[j][i]
            yt[1] = traj_y[j][i + 1]
            xo[0] = X[j][i]
            xo[1] = X[j][i + 1]
        elif i == N - 1:
            xt[0] = traj_x[j][i - 1]
            xt[1] = traj_x[j][i]
            yt[0] = traj_y[j][i - 1]
            yt[1] = traj_y[j][i]
            xo[0] = X[j][i - 1]
            xo[1] = X[j][i]
        else:
            xt[0] = traj_x[j][i - 1]
            xt[1] = traj_x[j][i + 1]
            yt[0] = traj_y[j][i - 1]
            yt[1] = traj_y[j][i + 1]
            xo[0] = X[j][i - 1]
            xo[1] = X[j][i + 1]

        if j == 0:
            xt[2] = traj_x[j][i]
            xt[3] = traj_x[j + 1][i]
            yt[2] = traj_y[j][i]
            yt[3] = traj_y[j + 1][i]
            yo[0] = Y[j][i]
            yo[1] = Y[j + 1][i]
        elif j == N - 1:
            xt[2] = traj_x[j - 1][i]
            xt[3] = traj_x[j][i]
            yt[2] = traj_y[j - 1][i]
            yt[3] = traj_y[j][i]
            yo[0] = Y[j - 1][i]
            yo[1] = Y[j][i]
        else:
            xt[2] = traj_x[j - 1][i]
            xt[3] = traj_x[j + 1][i]
            yt[2] = traj_y[j - 1][i]
            yt[3] = traj_y[j + 1][i]
            yo[0] = Y[j - 1][i]
            yo[1] = Y[j + 1][i]

        lambdas = eigs(xt, yt, xo, yo)
        if np.isnan(lambdas).all():
            ftle[j][i] = float("nan")
        else:
            ftle[j][i] = 0.5 * np.log(max(lambdas)) / (integration_time)

    return ftle


def FTLE(DT: DynTrack, integration_time: float, delta_t: float, copy: bool = False):
    """\
    Generate a scalar FTLE field from vector data.

    Parameters
    ----------
    DT
        A :class:`dyntrack.DynTrack` object.
    integration_time
        Overall integration time for sampling the vector field.
    delta_t
        Delta t used during the integration.
    copy
        Return a copy instead of writing to DT.

    Returns
    -------
    DT : :class:`dyntrack.DynTrack`
        if `copy=True` it returns or else add fields to `DT`:

        `.ftle`
            FTLE scalar values of the vector field.

    """

    DT = DT.copy() if copy else DT
    logg.info(
        "Obtaining FTLE scalar field (integration time: %s, delta t: %s)"
        % (integration_time, delta_t),
        reset=True,
    )
    traj_x, traj_y = get_traj(DT.X, DT.Y, DT.u, DT.v, integration_time, delta_t)
    logg.info("    Calculating FTLE scalar field", end="... ")
    ftle = get_ftle(traj_x, traj_y, DT.X, DT.Y, integration_time)
    logg.info("done")
    DT.ftle = ftle

    logg.info("    finished", time=True, end=" " if settings.verbosity > 2 else "\n")
    logg.hint("added \n" "    .ftle, FTLE scalar values of the vector field.")

    return DT if copy else None



import os
import pathlib
from setuptools import setup, find_packages
import subprocess
from os import path
from setuptools.dist import Distribution

dir = path.abspath(path.dirname(__file__))

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

with open(path.join(dir, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

from distutils.command.build import build


class CustomBuild(build):
    def run(self):
        build.run(self)
        subprocess.call(["make", "-C", "vfkm"])


class BinaryDistribution(Distribution):
    """Distribution which always forces a binary package with platform name"""

    def has_ext_modules(foo):
        return True


from setuptools.command.install import install


class InstallPlatlib(install):
    def finalize_options(self):
        install.finalize_options(self)
        if self.distribution.has_ext_modules():
            self.install_lib = self.install_platlib


def set_version():
    head = open(path.join(dir, ".git", "HEAD")).readline()
    if head.find("dev") != -1:
        return {
            "template": "{tag}.dev{ccount}",
            "dev_template": "{tag}.dev{ccount}",
            "dirty_template": "{tag}.dev{ccount}",
        }
    else:
        return {"template": "{tag}", "dev_template": "{tag}", "dirty_template": "{tag}"}


setup(
    name="dyntrack",
    version_config=set_version(),
    setup_requires=["setuptools-git-versioning"],
    description="Python package for the study of particle dynamics from 2D tracks",
    long_description=long_description,
    long_description_content_type="text/markdown",
    package_dir={"dyntrack": "dyntrack"},
    url="https://github.com/LouisFaure/DynTrack",
    packages=find_packages(),
    include_package_data=False,
    install_requires=requirements,
    cmdclass={"build": CustomBuild, "install": InstallPlatlib},
    distclass=BinaryDistribution,
)



from typing import Sequence, Union
from tqdm import tqdm
import warnings
import sys
import simpleppt
from .. import logging as logg
from .. import settings
from ..DynTrack import DynTrack

simpleppt.settings.verbosity = 0


def fit_ppt(
    DT: DynTrack,
    times: Union[Sequence, None] = None,
    lam: float = 10,
    sigma: float = 10,
    copy: bool = False,
    **kwargs,
):

    """\
    Compute a principal tree from partical position at each frame.

    Parameters
    ----------
    DT
        A :class:`dyntrack.DynTrack` object.
    times
        set of timepoints/frames to use for the fitting, by default uses all.
    lam
        Lambda parameter from SimplePPT algorithm.
    sigma
        Sigma parameter from SimplePPT algorithm.
    copy
        Return a copy instead of writing to DT.
    **kwargs
        Additional parameters to be passed to :func:`simpleppt.ppt`

    Returns
    -------
    DT : :class:`dyntrack.DynTrack`
        if `copy=True` it returns or else add fields to `DT`:

        `.ppts`
            List of principal trees calculated per frame.

    """

    logg.info(
        f"Generating principal tree per timepoints (lambda: {lam}, sigma: {sigma})",
        reset=True,
    )
    DT = DT.copy() if copy else DT

    tdata = DT.track_data

    times = tdata.Time.unique() if times is None else times
    ppts = []
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        for t in tqdm(times, file=sys.stdout, desc="    fitting"):
            tofit = tdata.loc[tdata.Time == t, ["Position X", "Position Y"]]
            ppts = ppts + [
                simpleppt.ppt(
                    tofit,
                    Nodes=tofit.shape[0],
                    lam=lam,
                    sigma=sigma,
                    progress=False,
                    **kwargs,
                )
            ]

    DT.ppts = ppts

    logg.info("    finished", time=True, end=" " if settings.verbosity > 2 else "\n")
    logg.hint("added \n" "    .ppts, list of SimplePPT objects.")

    return DT if copy else None



from typing import Union
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy as np
import warnings
from ..DynTrack import DynTrack
from matplotlib.axes import Axes


def vector_field(
    DT: DynTrack,
    density: float = 2,
    linewidth: float = 1,
    arrowsize: float = 1,
    arrowstyle="->",
    cmap="gnuplot",
    figsize: tuple = (7, 4),
    show: bool = True,
    **kwargs
):
    """\
    Plotting counterpart of `tl.vector_field`.

    Parameters
    ----------
    DT
        A :class:`dyntrack.DynTrack` object.
    cmap
        Colormap used by :func:`matplotlib.pyplot.countourf`.
    density
        Density of arrows used by :func:`matplotlib.pyplot.streamplot`.
    linewidth
        Arrow line width used by :func:`matplotlib.pyplot.streamplot`.
    arrowsize
        Arrow size used by :func:`matplotlib.pyplot.streamplot`.
    arrowstyle
        Arrow style used by :func:`matplotlib.pyplot.streamplot`.
    color
        Arrow color used by :func:`matplotlib.pyplot.streamplot`.
    figsize
        Figure size.
    show
        Show the plot, do not return axis.
    save
        Save plot to file
    **kwargs
        Arguments passed to :func:`matplotlib.pyplot.streamplot`.

    Returns
    -------
    :class:`matplotlib.axes.Axes` if `show=True`

    """

    fig, (ax, cax) = plt.subplots(
        1, 2, gridspec_kw={"width_ratios": [96, 4], "wspace": 0}
    )
    ax.set_aspect("equal")

    if DT.img is not None:
        ax.imshow(DT.img, origin="lower")
    speed = np.sqrt(DT.u ** 2 + DT.v ** 2)
    h = ax.streamplot(
        DT.X,
        DT.Y,
        DT.u,
        DT.v,
        color=speed,
        density=density,
        linewidth=linewidth,
        arrowsize=arrowsize,
        arrowstyle=arrowstyle,
        cmap=cmap,
        **kwargs
    )
    ax.axis("off")

    cbar = plt.colorbar(
        h.lines, ax=cax, ticks=[speed.min(), speed.max()], pad=0, fraction=1
    )
    cbar.ax.set_yticklabels(["low", "high"])
    cbar.set_label("particle speed", fontsize=12)
    cax.axis("off")

    if show == False:
        return (ax, cax)
    else:
        plt.show()
